//+------------------------------------------------------------------+
//|                                  NeuralNetworkClassification.mqh |
//|                                  Copyright 2023, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| neural network for classification model                          |
//+------------------------------------------------------------------+
class CNeuralNetwork
  {
protected:
   int               m_layers;                           // layers count
   ENUM_ACTIVATION_FUNCTION m_activation_functions[];    // layers activation functions
   matrix            m_weights[];                        // weights matrices between layers
   vector            m_values[];                         // layer values
   vector            m_values_a[];                       // layer values after activation
   vector            m_errors[];                         // error values for back propagation

public:
   //+------------------------------------------------------------------+
   //| Constructor                                                      |
   //+------------------------------------------------------------------+
   CNeuralNetwork(const int layers,const int& layer_sizes[],const ENUM_ACTIVATION_FUNCTION& act_functions[])
     {
      m_layers=layers;
      ArrayCopy(m_activation_functions,act_functions);
      ArrayResize(m_weights,layers-1);
      ArrayResize(m_values,layers);
      ArrayResize(m_values_a,layers);
      ArrayResize(m_errors,layers);

      //--- init weight matrices with uniform random
      for(int i=0; i<m_layers-1; i++)
        {
         double div=32767.0 * layer_sizes[i];  // divider

         m_weights[i].Init(layer_sizes[i],layer_sizes[i+1]);
         for(ulong j=0; j<m_weights[i].Rows()*m_weights[i].Cols(); j++)
            m_weights[i].Flat(j,rand()/div);
        }
     }

   //+------------------------------------------------------------------+
   //| Forward pass                                                                 |
   //+------------------------------------------------------------------+
   void ForwardFeed(const vector& input_data,vector& vector_pred)
     {
      m_values_a[0]=input_data;

      for(int i=0; i<m_layers-1; i++)
        {
         m_values[i+1]=m_values_a[i].MatMul(m_weights[i]);
         m_values[i+1].Activation(m_values_a[i+1],m_activation_functions[i+1]);
        }

      //--- predicted class probabilities
      vector_pred=m_values_a[m_layers-1];
     }

   //+------------------------------------------------------------------+
   //| error back propagation                                           |
   //+------------------------------------------------------------------+
   void BackPropagation(const vector& vector_true,double learning_rate)
     {
      //--- categorical cross entropy loss gradient
      m_errors[m_layers-1]=vector_true-m_values_a[m_layers-1];

      for(int i=m_layers-1; i>0; i--)
         m_errors[i-1].GeMM(m_errors[i],m_weights[i-1],1,0,TRANSP_B);

      //--- weights update
      for(int i=m_layers-1; i>0; i--)
        {
         vector back;
         m_values[i].Derivative(back,m_activation_functions[i]);
         back*=m_errors[i]*learning_rate;

         m_weights[i-1].GeMM(m_values_a[i-1],back,1,1);
        }
     }
  };
//+------------------------------------------------------------------+
